﻿using System;
using System.Linq;
using System.Runtime.InteropServices.WindowsRuntime;
using System.Threading.Tasks;
using Windows.Devices.Enumeration;
using Windows.Devices.SerialCommunication;
using Windows.Storage.Streams;

namespace RctcOS.Pololu
{
    /// <summary>
    /// This class implements some important methods from Pololu's Windows desktop library
    /// Usc.cs.
    /// The implementation is very trimmed down. Using Meastro Control Center, configure your Maestro in 
    /// the 'USB Dual Port' mode. 
    /// 
    /// This class used the Windows.Devices.SerialCommunication namespace to talk to the Meastro.
    /// You'll need to modify your Univeral app's manifest to allow those methods to work. Here's a
    /// sample element you'll need to add. This one is for the Maestro 12 channel device.
    //  <DeviceCapability Name = "serialcommunication" >
    ///      <Device Id="vidpid:1FFB 008A">
    ///          <Function Type = "name:serialPort" />
    ///      </Device>
    ///  </DeviceCapability>
    /// 
    /// </summary>

    public partial class Maestro : IDisposable
    {
        // This 'singleton' will fail across app domains. It's up to the user to make sure a Maestro
        // object is not instantiated in two different programs simultaneously, for example.
        private static Maestro _instance = null;
        public int Channels { get; private set; }
        public MaestroModel Model { get; private set; }
        private SerialDevice _serialDevice = null;
        public ushort LastError { get; private set; }

        private Maestro()
        {
            ;
        }

        /// <summary>
        /// Use this method to create an instance of a Maestro object and use it
        /// to communicate with the Maestro.
        /// </summary>
        /// <param name="model"></param>
        /// <returns></returns>
        public static async Task<Maestro> CreateMaestroController( MaestroModel model )
        {
            if( _instance == null )
            {
                ushort usbVid = Maestro.PololuUsbVid;
                ushort usbPid;
                int channels = 0;

                switch ( model )
                {
                    case MaestroModel.MicroMaestro6Channel:
                        usbPid = Maestro.MicroMaestro6ChannelPid;
                        channels = 6;
                        break;

                    case MaestroModel.MiniMaestro12Channel:
                        usbPid = Maestro.MiniMaestro12Channel;
                        channels = 12;
                        break;

                    case MaestroModel.MiniMaestro18Channel:
                        usbPid = Maestro.MiniMaestro18Channel;
                        channels = 18;
                        break;

                    case MaestroModel.MiniMaestro24Channel:
                        usbPid = Maestro.MiniMaestro24Channel;
                        channels = 24;
                        break;

                    default:
                        throw new Exception($"Invalid model {model}");
                }

                _instance = new Maestro
                {
                    Model = model,
                    Channels = channels
                };

                string aqs = Windows.Devices.SerialCommunication.SerialDevice.GetDeviceSelectorFromUsbVidPid(usbVid, usbPid);
                DeviceInformationCollection interfaces = await DeviceInformation.FindAllAsync(aqs, null);

                // We want to find the USB command port interface.
                DeviceInformation deviceInterface = interfaces.FirstOrDefault(i => i.Id.ToLower().Contains("mi_00"));
                if(deviceInterface == null )
                {
                    throw new Exception($"Cannot find the Usb interface for the Maestro device.");
                }

                try
                {
                    _instance._serialDevice = await SerialDevice.FromIdAsync(deviceInterface.Id);
                    // I'm not sure these serial parameters are necessary. Better safe than sorry.
                    _instance._serialDevice.StopBits = SerialStopBitCount.One;
                    _instance._serialDevice.Parity = SerialParity.None;
                    _instance._serialDevice.DataBits = 8;

                    //await SendPolouServoCmd(_serialDevice, 0, 1200);

                    //await Task.Delay(1000);

                    //await SendPolouServoCmd(_serialDevice, 0, 1700);

                }

                catch (Exception ex)
                {
                    throw new Exception($"Cannot find create a serial device interface for the Maestro device. Check the app manifest for the 'serialcommunication' DeviceCapability", ex);
                }

                return (_instance);
            }
            else
            {
                throw new Exception("Maestro device is already open");
            }
        }

        /// <summary>
        /// Set the position of the specified servo. See the "Pololu Maestro Servo Controller User’s Guide"
        /// chapter 5.e for more information.
        /// </summary>
        /// <param name="servo">Servos are numbered from zero.</param>
        /// <param name="microSeconds">The position to set the servo. Most hobby servos have a range
        /// of 1000 (fully left) to 2000 (fully right)</param>
        /// <returns></returns>
        public async Task SetTarget(byte servo, ushort microSeconds)
        {
            try
            {
                // The Maestro wants the position specified in 1/4 microseconds.
                int quarterMicroSeconds = microSeconds * 4;
                IBuffer iBuffer = WindowsRuntimeBufferExtensions.AsBuffer(
                    new byte[] { (byte)Command.SetTarget, servo, (byte)(quarterMicroSeconds & 0x7F), (byte)((quarterMicroSeconds >> 7) & 0x7F) });

                await _instance._serialDevice.OutputStream.WriteAsync(iBuffer);
                await GetPololuErrors();
            }

            catch (Exception e)
            {
                throw new Exception( $"Failed to set target of servo {servo} to {microSeconds}.", e);
            }
        }

        /// <summary>
        /// Sets the speed of the servo. See the "Pololu Maestro Servo Controller User’s Guide"
        /// chapter 5.e for more information.
        /// </summary>
        /// <param name="servo">Servos are numbered from zero.</param>
        /// <param name="speed"></param>
        /// <returns></returns>
        public async Task SetSpeed(byte servo, ushort speed)
        {
            try
            {
                IBuffer iBuffer = WindowsRuntimeBufferExtensions.AsBuffer(
                    new byte[] { (byte)Command.SetSpeed, servo, (byte)(speed & 0x7F), (byte)((speed >> 7) & 0x7F) });

                await _instance._serialDevice.OutputStream.WriteAsync(iBuffer);
                await GetPololuErrors();
            }
            catch (Exception e)
            {
                throw new Exception( $"Failed to set speed of servo {servo} to {speed}.", e);
            }
        }

        /// <summary>
        /// Sets the specified servo acceleration. See the "Pololu Maestro Servo Controller User’s Guide"
        /// chapter 5.e for more information.
        /// </summary>
        /// <param name="servo">Servos are numbered from zero.</param>
        /// <param name="acceleration"></param>
        /// <returns></returns>
        public async Task SetAcceleration(byte servo, ushort acceleration)
        {
            // set the high bit of servo to specify acceleration
            try
            {
                IBuffer iBuffer = WindowsRuntimeBufferExtensions.AsBuffer(
                    new byte[] { (byte)Command.SetAcceleration, servo, (byte)(acceleration & 0x7F), (byte)((acceleration >> 7) & 0x7F) });

                await _instance._serialDevice.OutputStream.WriteAsync(iBuffer);
                await GetPololuErrors();
            }
            catch (Exception e)
            {
                throw new Exception( $"Failed to set acceleration of servo {servo} to {acceleration}.", e);
            }
        }

        /// <summary>
        /// Returns the position of the specified servo in micorseconds.
        /// </summary>
        /// <param name="servo">Servos are numbered from zero.</param>
        /// <returns></returns>
        public async Task<ushort> GetPosition(byte servo)
        {
            // set the high bit of servo to specify acceleration
            try
            {
                IBuffer iBuffer = WindowsRuntimeBufferExtensions.AsBuffer(
                    new byte[] { (byte)Command.GetPosition, servo });

                await _instance._serialDevice.OutputStream.WriteAsync(iBuffer);
                Windows.Storage.Streams.Buffer inBuf = new Windows.Storage.Streams.Buffer(2);
                await _instance._serialDevice.InputStream.ReadAsync(inBuf, 2, InputStreamOptions.None);
                byte b0 = inBuf.GetByte(0);
                byte b1 = inBuf.GetByte(1);

                ushort quarterMicroseconds = (ushort)(b1 * 256 + b0);
                ushort microSeconds = (ushort)(quarterMicroseconds / 4);
                return (microSeconds);
            }
            catch (Exception e)
            {
                throw new Exception($"Failed to get position of servo {servo}.", e);
            }
        }

        /// <summary>
        /// Queries the Maestro to see of any servos are currently moving.
        /// </summary>
        /// <returns>True if any servos are moving, false otherwise.</returns>
        public async Task<bool> AnyServosMoving()
        {
            if( _instance.Model == MaestroModel.MicroMaestro6Channel )
            {
                throw new Exception("The 'Get Moving State' state command is not compatible with the MicroMaestro");
            }

            // set the high bit of servo to specify acceleration
            try
            {
                IBuffer iBuffer = WindowsRuntimeBufferExtensions.AsBuffer(
                    new byte[] { (byte)Command.GetMovingState });

                await _instance._serialDevice.OutputStream.WriteAsync(iBuffer);
                Windows.Storage.Streams.Buffer inBuf = new Windows.Storage.Streams.Buffer(1);
                await _instance._serialDevice.InputStream.ReadAsync(inBuf, 2, InputStreamOptions.None);
                byte b0 = inBuf.GetByte(0);
                return (b0 == (ushort)1);
            }
            catch (Exception e)
            {
                throw new Exception($"Failed to get moving state of servos.", e);
            }
        }

        private async Task GetPololuErrors()
        {
            IBuffer oBuffer = WindowsRuntimeBufferExtensions.AsBuffer(new byte[] { (byte)Command.GetErrors });
            await _instance._serialDevice.OutputStream.WriteAsync(oBuffer);

            Windows.Storage.Streams.Buffer inBuf = new Windows.Storage.Streams.Buffer(2);
            await _instance._serialDevice.InputStream.ReadAsync(inBuf, 2, InputStreamOptions.None);
            byte b0 = inBuf.GetByte(0);
            byte b1 = inBuf.GetByte(1);
            _instance.LastError =(ushort)(b1 * (ushort)256 + b0);
        }

        #region IDisposable Support
        private bool disposedValue = false; // To detect redundant calls

        protected virtual void Dispose(bool disposing)
        {
            if (!disposedValue)
            {
                if (disposing)
                {
                    if( _instance?._serialDevice != null )
                    {
                        _instance._serialDevice.Dispose();
                    }
                }

                if (_instance?._serialDevice != null)
                {
                    _instance._serialDevice = null;
                    _instance = null;
                }

                disposedValue = true;
            }
        }

        // TODO: override a finalizer only if Dispose(bool disposing) above has code to free unmanaged resources.
        // ~Maestro() {
        //   // Do not change this code. Put cleanup code in Dispose(bool disposing) above.
        //   Dispose(false);
        // }

        // This code added to correctly implement the disposable pattern.
        public void Dispose()
        {
            // Do not change this code. Put cleanup code in Dispose(bool disposing) above.
            Dispose(true);
            // TODO: uncomment the following line if the finalizer is overridden above.
            // GC.SuppressFinalize(this);
        }
        #endregion
    }
}
